#!/usr/bin/env python3

import argparse
import subprocess
import os
import sys

version = "3.0.0"


def create_config(configfile, args):
    # check existence
    if not os.path.exists(args.sample_list):
        print("path to sample list " + args.sample_list + "does not exist")
        sys.exit(1)

    if not os.path.exists(args.cluster_names):
        print("path to cluster_names " + args.cluster_names + "does not exist")
        sys.exit(1)

    if not os.path.exists(args.external_cluster_names):
        print("path to external_cluster_names " + args.external_cluster_names + "does not exist")
        sys.exit(1)

    # check optional input
    if os.stat(args.serovar_info).st_size == 0:
        print("No serovar file provided")
        do_serovar_info = False
    else:
        do_serovar_info = True

    #  initialize sample2cluster file if necessary
    mapping_sample2clustername_file = os.path.join(args.working_directory, "merged_db",
                                                   "mapping_sample2clustername.tsv")
    if not os.path.exists(mapping_sample2clustername_file):
        print("Initializing file" + mapping_sample2clustername_file)
        if not os.path.exists(os.path.dirname(mapping_sample2clustername_file)):
            try:
                os.makedirs(os.path.dirname(mapping_sample2clustername_file))
            except:
                print("Directory merged_db could NOT be created")
                sys.exit(1)
        with open(mapping_sample2clustername_file, "w") as f:
            bytereturn = f.write("sample\tcluster_name\n")


    outfile = open(configfile, "w")
    outfile.write("#Created by chewieSnake_join.py v{}\n".format(version))

    outfile.write('workdir:\n    "{}"\n'.format(args.working_directory))
    outfile.write('samples:\n    "{}"\n'.format(args.sample_list))
    outfile.write(
        'parameters:\n  clustering_method: "{}"\n  distance_threshold: {}\n  grapetree_distance_method: {}\n  cluster_names_reservoir: {}\n  subcluster_distance_thresholds: {}\n  subcluster_name_types: {}\n  external_cluster_names: {}\n  serovar_info: {}\n  do_serovar_info: {}\n  species_shortname: "{}"\n'.format(
            args.clustering_method, args.distance_threshold, args.distance_method, args.cluster_names,
            args.subcluster_thresholds, args.subcluster_names, args.external_cluster_names, args.serovar_info,
            do_serovar_info, args.species_shortname))

    outfile.close()


def run_snakemake(configfile, args):
    if args.forceall:
        forceall = "--forceall"
    else:
        forceall = ""

    if args.unlock:
        unlock = "--unlock"
    else:
        unlock = ""

        # Change this if more all rules are availalbe
    if args.noreport:
        force = "-f all_noreport"
    elif args.cluster:
        force = "-f all_clustering"
    elif args.report_only:
        force = "-f all_reportonly"
    else:
        force = ""

    if args.dryrun:
        dryrun = "-n"
    else:
        dryrun = ""

    if args.conda_frontend:
        frontend = "conda"
    else:
        frontend = "mamba"

    if args.use_conda:
        call = "snakemake -p --conda-prefix {CondaPrefix} --use-conda --keep-going --configfile {configfile} --snakefile {snakefile} {force} {dryrun} {forceall} {unlock} --cores {threads} --conda-frontend {frontend}".format(
            CondaPrefix=args.condaprefix, snakefile=args.snakefile, configfile=configfile, force=force, dryrun=dryrun,
            forceall=forceall, unlock=unlock, threads=args.threads, frontend=frontend)
    else:
        call = "snakemake -p --keep-going --configfile {configfile} --snakefile {snakefile} {force} {dryrun} {forceall} {unlock} --cores {threads}".format(
            CondaPrefix=args.condaprefix, snakefile=args.snakefile, configfile=configfile, force=force, dryrun=dryrun,
            forceall=forceall, unlock=unlock, threads=args.threads)

    print(call)
    subprocess.call(call, shell=True)
    # call again for checkpoints # TODO only if not an error
    if not args.dryrun:
        subprocess.call(call, shell=True)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--sample_list',
                        help='List of allele profiles to analyze, containg a tab separated file with the columns: sender profiles statistics timestamps',
                        required=True, type=os.path.abspath)
    parser.add_argument('-d', '--working_directory', help='Working directory where results are saved', required=True,
                        type=os.path.abspath)
    parser.add_argument('-s', '--snakefile',
                        help='Path to Snakefile of chewieSnake pipeline, default is path to Snakefile in same directory',
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "chewieSnake_join.smk"),
                        required=False)
    parser.add_argument('--condaprefix',
                        help='Path of default conda environment, enables recycling built environments, default is in folder conda_env in repository directory.',
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "conda_env"), required=False)

    # clustering parameters
    parser.add_argument('--clustering_method',
                        help=' The agglomeration method to be used for hierarchical clustering. This should be (an unambiguous abbreviation of) one of "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC) or "centroid" (= UPGMC); default = single',
                        default="single", required=False)
    parser.add_argument('--distance_threshold',
                        help='A single distance threshold for the exctraction of sub-trees; default = 10', default=10,
                        required=False)
    parser.add_argument('--cluster_names',
                        help='A file with potential names for cluster names, one name per line, default: repo_path/scripts/cluster_names_reservoir.txt',
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "scripts",
                                             "cluster_names_reservoir.txt"), required=False)
    parser.add_argument('--subcluster_thresholds',
                        help='A list of distance thresholds for subclustering; default = [3]', default=[3],
                        required=False)
    parser.add_argument('--subcluster_names',
                        help='A list of name types for nomenclature of subclusters; choose from "numerical" and "greek-alphabet"; one entry for each subcluster_threshold; default = ["greek-alphabet"]',
                        default=["greek-alphabet"], required=False)
    parser.add_argument('-e', '--external_cluster_names',
                        help='A tab seperated file with external cluster names, e.g. Epi clusters. Must contain one representative sample per cluster: sample<tab>cluster_name ',
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "scripts",
                                             "epi_clusters.tsv"), required=False, type=os.path.abspath)
    parser.add_argument('--serovar_info',
                        help='A tab seperated file with serovar or clade names: sample<tab>cluster_name ',
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "scripts",
                                             "sample2serovar.tsv"), required=False, type=os.path.abspath)
    # TODO add here, in config and as rule parser.add_argument('--address_range', help='A comma separated set of cutoff values for defining the clustering address (Default: 1,5,10,20,30,50,100,200,500,990,1000)', default="1,5,10,20,30,50,100,200,500,990,1000", required=False)

    parser.add_argument('--noreport', help='Do not create html report', default=False, action='store_true',
                        required=False)
    parser.add_argument('--report_only', help='Only create html report', default=False, action='store_true',
                        required=False)
    parser.add_argument('--cluster', help='Assign cluster type and cluster addresses', default=False,
                        action='store_true', required=False)
    parser.add_argument('--distance_method', help='Grapetree distance method; default = 3', default=3, required=False)
    parser.add_argument('--species_shortname', help='Abbreviated Species name for cluster names; default = "SE"',
                        default="SE", required=False)
    parser.add_argument('--use_conda',
                        help='Utilize "--useconda" option, i.e. all software dependencies are available in a single env',
                        default=False, action='store_true',
                        required=False)
    parser.add_argument('--conda_frontend',
                        help='Do not mamba but conda as frontend to create individual module\' software environments',
                        default=False, action='store_true',
                        required=False)
    parser.add_argument('-t', '--threads',
                        help='Number of Threads to use. Note that samples can only be processed sequentially due to the required database access. However the allele calling can be executed in parallel for the different loci, default = 10',
                        default=10, required=False)
    parser.add_argument('-n', '--dryrun', help='Snakemake dryrun. Only calculate graph without executing anything',
                        default=False, action='store_true', required=False)
    parser.add_argument('-f', '--force', help='Snakemake force. Force recalculation of all steps', default=False,
                        action='store_true', required=False)
    parser.add_argument('--forceall', help='Snakemake force. Force recalculation of all steps', default=False,
                        action='store_true', required=False)
    parser.add_argument('--unlock', help='unlock snakemake', default=False, action='store_true', required=False)
    parser.add_argument('--logdir', help='Path to directory whete .snakemake output is to be saved', default='NA',
                        required=False)

    args = parser.parse_args()

    if not os.path.exists(args.working_directory):
        os.makedirs(args.working_directory)

    configfile = os.path.join(args.working_directory, "config.yaml")
    create_config(configfile, args)

    if os.path.exists(configfile):
        run_snakemake(configfile, args)
    else:
        print("Path to configfile does not exist: " + configfile)


if __name__ == '__main__':
    main()
