# cgmlst_chewbbaca snakemake workflow: Input is a set of reports

import os
import pandas as pd
import shutil
shell.executable("bash")


# Settings ------------------------------------------------------------


# Set snakemake main workdir variable

workdir: config["workdir"]

samples = pd.read_csv(config["samples"], index_col="sender", sep = "\t")
samples.index = samples.index.astype('str', copy=False) # in case samples are integers, need to convert them to str


# Functions ------------------------------------------------------------

def _get_profile(wildcards,file='profiles'):
    return samples.loc[(wildcards.sender), [file]].dropna()[0]

def _get_assembly(wildcards,file='assembly'):
    return samples.loc[(wildcards.sample), [file]].dropna()[0]


# input

# rules ------------------------------------------------------------


# only merging, no report
rule all_merge:
    input:
        "merged_db/allele_profiles.tsv",
        "merged_db/allele_statistics.tsv",
        "merged_db/sender_sample_info.tsv",
        "merged_db/distance_matrix.tsv",
        "merged_db/timestamps.tsv",
        "merged_db/orphan_samples.tsv", # from rule clustering
        "merged_db/listofreports.txt", # result from checkpoint
        "merged_db/report.html"


rule all_noreport:
    input:
        "merged_db/allele_profiles.tsv",
        "merged_db/allele_statistics.tsv",
        "merged_db/sender_sample_info.tsv",
        "merged_db/distance_matrix.tsv",
        "merged_db/timestamps.tsv",


rule all_clustering:
    input:
        "merged_db/allele_profiles.tsv",
        "merged_db/allele_statistics.tsv",
        "merged_db/sender_sample_info.tsv",
        "merged_db/distance_matrix.tsv",
        "merged_db/timestamps.tsv",
        "merged_db/orphan_samples.tsv", # from rule clustering
        #"merged_db/samples2clusters.csv" # result from checkpoint
        "merged_db/listofreports.txt", # result from checkpoint
        "merged_db/report.html"

rule all_reportonly:
    input:
        "merged_db/report.html"



# ------------


# rule is necessary since the lambda function cannot be expanded
rule copy_profiles: 
    input:
        profiles = lambda wildcards: _get_profile(wildcards, 'profiles')
    output:
        merged_db = temporary("merged_db/tmp/allele_profiles_{sender}.tsv")
    message: "Copying all allele profiles"
    #log: 
    shell:
        """
        cat {input.profiles} | sed 's/^FILE/#FILE/' > {output}
        """

rule join_profiles:
    input:
        profiles = expand("merged_db/tmp/allele_profiles_{sender}.tsv", sender = samples.index)
    output:
        profiles = "merged_db/allele_profiles.tsv"
    message: "Merging all allele profiles"
    run:
        with open(output.profiles, 'wb') as outfile:
            for i, fname in enumerate(input.profiles):
                with open(fname, 'rb') as infile:
                    if i != 0:
                        infile.readline()  # Throw away header on all but first file
                    shutil.copyfileobj(infile, outfile)      # Block copy rest of file from input to output without parsing

rule sample_sender_relation: 
    input:
        profiles = "merged_db/tmp/allele_profiles_{sender}.tsv"
    output:
        info = temporary("merged_db/tmp/sender_sample_{sender}.tsv")
    message: "Extracting sender sample info"
    shell:
        """
        awk -F ' ' -v sender={wildcards.sender} '{{print $1"\t"sender}}' {input.profiles} > {output}
        """


rule sample_sender_relation_merge: 
    input:
        info = expand("merged_db/tmp/sender_sample_{sender}.tsv", sender=samples.index)
    output:
        info = "merged_db/sender_sample_info.tsv"
    message: "Merging all sender info files"
    run:
        with open(output.info, 'wb') as outfile:
            for i, fname in enumerate(input.info):
                with open(fname, 'rb') as infile:
                    if i != 0:
                        infile.readline()  # Throw away header on all but first file
                    shutil.copyfileobj(infile, outfile)      # Block copy rest of file from input to output without parsing

#rule join_statistics:

# rule is necessary since the lambda function cannot be expanded
rule copy_statistics: 
    input:
        profiles = lambda wildcards: _get_profile(wildcards, 'statistics')
    output:
        merged_db = temporary("merged_db/tmp/allele_statistics_{sender}.tsv")
    message: "Copying all allele statistics"
    #log: 
    shell:
        """
        cat {input.profiles} > {output}
        """

rule join_statistics:
    input:
        partner_stats = expand("merged_db/tmp/allele_statistics_{sender}.tsv", sender = samples.index)
    output:
        merged_stats = "merged_db/allele_statistics.tsv"
    message: "Merging all allele statistics"
    run:
        with open(output.merged_stats, 'wb') as outfile:
            for i, fname in enumerate(input.partner_stats):
                with open(fname, 'rb') as infile:
                    if i != 0:
                        infile.readline()  # Throw away header on all but first file
                    shutil.copyfileobj(infile, outfile)      # Block copy rest of file from input to output without parsing



# rule: join timestamps
rule copy_timestamps: 
    input:
        profiles = lambda wildcards: _get_profile(wildcards, 'timestamps')
    output:
        merged_db = temporary("merged_db/tmp/timestamps_{sender}.tsv")
    message: "Copying all timestamps"
    shell:
        """
        cat {input.profiles} > {output}
        """

rule join_timestamps:
    input:
        profiles = expand("merged_db/tmp/timestamps_{sender}.tsv", sender = samples.index)
    output:
        merged_timestamps = "merged_db/timestamps.tsv"
    message: "Merging all timestamps"
    shell:
        "cat {input.profiles} | grep -Pv '^sample\tdate' > {output}"



## Re-use rules from here on ---------
rule run_grapetree_merged:
    input:
        merged_db = "merged_db/allele_profiles.tsv"
    output:
        distance_matrix_tsv = "merged_db/distance_matrix.tsv",
        mstree = "merged_db/grapetree_mstree.tre",
    conda: "envs/grapetree3.yaml"
    message:
      "Running rule grapetree"
    params:
        distance_option = config["parameters"]["grapetree_distance_method"], #
    shell:
        """
        grapetree -p {input} -m distance --missing {params.distance_option} | sed -E 's/[[:space:]]{{2,}}/ /' | tail -n+2 > {output.distance_matrix_tsv}
        grapetree -p {input} -m MSTreeV2 > {output.mstree}
        """
# -----------



checkpoint clustering:
    input:
        distance_matrix_tsv = "merged_db/distance_matrix.tsv",
        mapping_sample2clustername_file = "merged_db/mapping_sample2clustername.tsv"
    output:
        clustering_dir = directory("merged_db/clustering"),
        orphansamples_file = "merged_db/orphan_samples.tsv"
    message: "Clustering of distance matrix"
    log: "logs/clustering.log"
    params:
        functions_file = {workflow.basedir},
        mapping_sample2clustername_file = "merged_db/mapping_sample2clustername.tsv",
        distance_threshold = config["parameters"]["distance_threshold"], # 20
        clustering_method = config["parameters"]["clustering_method"], #"average",
        potential_names_file = config["parameters"]["cluster_names_reservoir"]
    conda: "envs/rmarkdown.yaml"
    script:
        "scripts/clustering.R"


# match epi cluster names; only perform if file exists; make this output an optional input for report rule
rule match_external_cluster_names:
    input:
        external_cluster_names = config["parameters"]["external_cluster_names"],
        mapping_sample2clustername_file = "merged_db/mapping_sample2clustername.tsv",
        orphansamples_file = "merged_db/orphan_samples.tsv", # not used, but sets order of rule
    output:
        mapping_sample2externalclustername_file = "merged_db/mapping_sample2externalclustername.tsv",
    message: "Matching externally provided cluster names to internal clusters"
    log: "logs/match_external_cluster_names.log"
    conda: "envs/rmarkdown.yaml"
    script:
        "scripts/match_clusternames.R"


rule match_serovar:
    input:
        external_cluster_names = config["parameters"]["serovar_info"],
        mapping_sample2clustername_file = "merged_db/mapping_sample2clustername.tsv",
        orphansamples_file = "merged_db/orphan_samples.tsv", # not used, but sets order of rule
    output:
        mapping_sample2externalclustername_file = "merged_db/mapping_sample2serovar.tsv",
    message: "Matching externally provided serovar names to internal clusters"
    log: "logs/match_serovar_names.log"
    script:
        "scripts/match_clusternames.R"


#sub_distance matrices
rule cluster_distance_matrix:
    input:
        samples_in_cluster = "merged_db/clustering/samples_{i}.tsv",
        distance_matrix_tsv = "merged_db/distance_matrix.tsv",
    output:
        distance_matrix_tsv = "merged_db/clustering/{i}/distance_matrix.tsv"
    log: "logs/cluster_distance_matrix_{i}.log"
    message: "Clustering of sub distance matrices"
    params:
        cluster_matrix = True,
        functions_file = {workflow.basedir},
    script:
        "scripts/cluster_distancematrix.R"


# ------------------
# persistent subclusters; make this optional ???

rule subclustering:
    input:
        distance_matrix_cluster_tsv = "merged_db/clustering/{i}/distance_matrix.tsv",
    output:
        subclusters = "merged_db/clustering/{i}/mapping_sample2subclustername.tsv"
    message: "Clustering of distance matrix for cluster"
    log: "logs/subclustering_{i}.log"
    params:
        functions_file = {workflow.basedir},
        mapping_sample2clustername_file = "merged_db/clustering/{i}/mapping_sample2subclustername.tsv",
        distance_thresholds = config["parameters"]["subcluster_distance_thresholds"], # [10,3], 
        name_types = config["parameters"]["subcluster_name_types"], # ["numerical","greek-alphabet"], # 
        clustering_method = config["parameters"]["clustering_method"], #"single",
    conda: "envs/rmarkdown.yaml"
    script:
        "scripts/subclustering.R"


# report per cluster ------------------

rule cluster_report:
    input:
        samples2cluster = "merged_db/clustering/samples_{i}.tsv",
        distance_matrix_tsv = "merged_db/clustering/{i}/distance_matrix.tsv",
        subclusters = "merged_db/clustering/{i}/mapping_sample2subclustername.tsv",
        mapping_sample2externalclustername_file = "merged_db/mapping_sample2externalclustername.tsv",
        sender_info = "merged_db/sender_sample_info.tsv",
        timestamps = "merged_db/timestamps.tsv",
    output:
        "merged_db/clustering/{i}/clusterreport.html"
    log: "logs/cluster_report_{i}.log"
    params:
        workdir = config["workdir"],
        cluster = "{i}", 
        export = True,
        distance_threshold = config["parameters"]["distance_threshold"],
        clustering_method = config["parameters"]["clustering_method"],
        mapping_sample2serovar = "merged_db/mapping_sample2serovar.tsv" if config["parameters"]["do_serovar_info"] else {}, # param or input?,
        functions_file = {workflow.basedir},
        species_shortname = config["parameters"]["species_shortname"] # "SE"
    conda: "envs/rmarkdown.yaml"
    message: "Creating Rmarkdown report for cluster"
    script:
        "scripts/cluster_report.Rmd"



# aggregate rules for collection of checkpoint dynamic files:

# important: expand results from intermediate files BUT glob_wildcards from checkpoint!!!
def aggregate_clusterreport(wildcards):
    checkpoint_output = checkpoints.clustering.get(**wildcards).output[0]
    return expand("merged_db/clustering/{i}/clusterreport.html",
            i=glob_wildcards(os.path.join(checkpoint_output, "samples_{i}.tsv")).i)


def aggregate_subclusters(wildcards):
    checkpoint_output = checkpoints.clustering.get(**wildcards).output[0]
    return expand("merged_db/clustering/{i}/mapping_sample2subclustername.tsv",
            i=glob_wildcards(os.path.join(checkpoint_output, "samples_{i}.tsv")).i)


# an aggregation over all produced clusters
rule aggregate_clusterreport:
    input:
        aggregate_clusterreport
    output:
        "merged_db/listofreports.txt"
    shell:
        """
        for file in {input}; do
            echo $file >> {output}
        done
        """

# -----------


rule gather_cluster_information:
    input:
        sender_info = "merged_db/sender_sample_info.tsv",
        timestamps = "merged_db/timestamps.tsv",
        mapping_sample2externalclustername_file = "merged_db/mapping_sample2externalclustername.tsv",
        mapping_sample2clustername_file = "merged_db/mapping_sample2clustername.tsv",
        mapping_sample2serovar = "merged_db/mapping_sample2serovar.tsv" if config["parameters"]["do_serovar_info"] else {}, # param or input?
        orphansamples_file = "merged_db/orphan_samples.tsv", # not used, but sets order of rule
        subclusters = aggregate_subclusters, # dynamic
    output:
        clustering_info = "merged_db/sample_cluster_information.tsv",
        orphan_info = "merged_db/sample_orphan_information.tsv"
    log: "logs/cluster_information.log"
    message: "Gathering cluster information"
    params:
        workdir = config["workdir"],
    script:
        "scripts/cluster_information.R"


rule report_merged:
    input:
        distance_matrix = "merged_db/distance_matrix.tsv",
        allele_stats = "merged_db/allele_statistics.tsv",
        nwk  = "merged_db/grapetree_mstree.tre",
        sender_info = "merged_db/sender_sample_info.tsv",
        timestamps = "merged_db/timestamps.tsv",
        report_links = "merged_db/listofreports.txt",
        mapping_sample2externalclustername_file = "merged_db/mapping_sample2externalclustername.tsv",
        clustering_info = "merged_db/sample_cluster_information.tsv",
        orphan_info = "merged_db/sample_orphan_information.tsv",
        cluster_report = aggregate_clusterreport
    output:
        "merged_db/report.html"
    log: "logs/rmarkdown_report.log"
    params:
        workdir = config["workdir"],
        exportpath = "merged_db/exported_trees",
        export = True,
        distance_threshold = config["parameters"]["distance_threshold"], # default = 30
        clustering_method = config["parameters"]["clustering_method"], # default = "single",
        mapping_sample2serovar = "merged_db/mapping_sample2serovar.tsv",
        sample2serovar = config["parameters"]["serovar_info"],
        parameters = config["parameters"],
        species_shortname = config["parameters"]["species_shortname"], # "SE"
        functions_file = {workflow.basedir},
    conda: "envs/rmarkdown.yaml"
    message: "Creating Rmarkdown report for all samples"
    script:
        "scripts/report_joined.Rmd"

# logging information---------


onstart:
    shell('echo -e "running\t`date`" > pipeline_status.txt')

onsuccess:
    shell('echo -e "success\t`date`" > pipeline_status.txt')

onerror:
    shell('echo -e "error\t`date`" > pipeline_status.txt'),
    print("An error occurred")

