#!/usr/bin/env Rscript

# Distance matrix reduce version 0.0.1
# Copyright (C) 2019 onwards Carlus Deneke

# Goal: Get closest neighbors

# supress warnings
options(warn=-1)

suppressPackageStartupMessages(library(magrittr))
suppressPackageStartupMessages(library("ape"))
suppressPackageStartupMessages(library(dendextend))
suppressPackageStartupMessages(library(reshape2))
suppressPackageStartupMessages(library(ggplot2))


options(warn=0)

# snakemake input and output -------------------------------------

# dev mode
dev_mode <- F # receive parameters from snakemake, set to T for testing purposes

if(dev_mode){
  source(file = "test_DistanceMatrix_reduce.R")
} else {

  distance_matrix_file <- file.path(snakemake@input[["distance_matrix"]])
  
  # params
  threshold <- snakemake@params[["threshold"]]
  external_name <- snakemake@params[["name"]]
  internal_name <- "comparison_db"
  outdir <- snakemake@params[["outdir"]]
  sample_list_file <- snakemake@params[["samplesheet"]]
  clustering_method <- snakemake@params[["clustering_method"]]
  
  # output ---
  samples_related_file <- snakemake@output[["samples_related"]]
  match_overview_file <- snakemake@output[["match_overview"]]
  distance_out <- snakemake@output[["distance_matrix_filtered"]]
  distance_table_out=file.path(outdir,"distance_table.tsv")
  distance_table_out2=file.path(outdir,"distance_table_transposed.tsv")
  sample_meta_file <- snakemake@output[["sample_meta"]]
  
  #distance_table_pdf=file.path(outdir,"distance_table.pdf")
  distance_matrix_pdf=file.path(outdir,"distance_matrix_filtered.pdf")
  tree_nwk=file.path(outdir,"tree_singlelinkage.nwk")
  tree_pdf=file.path(outdir,"tree_singlelinkage.pdf")
  tree_png=file.path(outdir,"tree_singlelinkage.png")
  
  # redirect output to log file
  log <- file(snakemake@log[[1]], open="wt")
  sink(log)
  sink(log, type="message")
  
  force <- T
  
  }


# check existence
if (!file.exists(distance_matrix_file)){
  stop(paste("File",distance_matrix_file,"does NOT exist"))
}

if (!file.exists(sample_list_file)){
  stop(paste("File",sample_list_file,"does NOT exist"))
}

if (!dir.exists(outdir)){
  dir.create(outdir)
} else {
  warning(paste("Output dir",outdir,"already exists."))
}

# print parameters -------------------------------------------------
print("Parameters:")
print(paste("distance_matrix_file:",distance_matrix_file))
print(paste("sample_list_file:",sample_list_file))
print(paste("external_name:",external_name))
print(paste("threshold:",threshold))
print(paste("outdir:",outdir))


if (file.exists(distance_out) && force == FALSE){
  stop(paste("File",distance_out,"already exists. Please specify --force to overwrite"))
}

# functions ----------------------------------

read_distance_matrix <- function(file, sep = " ", header = FALSE){
  if(!file.exists(file)) {stop(paste("File",file,"does Not exist")); return(NA)}
  distance_matrix <- read.delim(file, sep=sep, stringsAsFactors = F, header = header)
  rownames(distance_matrix) <- sub(".fasta","",distance_matrix[,1])
  distance_matrix <- distance_matrix[,-1]
  colnames(distance_matrix) <- rownames(distance_matrix)
  distance_matrix <- as.matrix(distance_matrix[order(colnames(distance_matrix)),order(colnames(distance_matrix))])
  return(distance_matrix)
}

reorder_cormat <- function(cormat){
  # Use correlation between variables as distance
  #dd <- as.dist((1-cormat)/2)
  if(nrow(cormat)>2){
    dd <- as.dist(cormat)
    hc <- hclust(dd)
    cormat <-cormat[hc$order, hc$order]  
  }
  return(cormat)
}

# re-order matrix
#snp_matrix_clustered <- hclust(as.dist(snp_matrix))
#snp_matrix <- snp_matrix[snp_matrix_clustered$order, snp_matrix_clustered$order]



get_upper_tri <- function(cormat){
  cormat[lower.tri(cormat)]<- NA
  return(cormat)
}

plot_distance_matrix <- function(distance_matrix2, param_color_thr = c(0,1,5,10,20), param_size = 3, param_legendsize = 12, metadata = NA, label_color = "black"){
  
  # remove diagonal elements
  diag(distance_matrix2) <- NA
  
  # Reorder the correlation matrix
  distance_matrix2 <- reorder_cormat(distance_matrix2)
  
  # convert to data frame
  
  # melt matrix
  distance_matrix.m <- reshape2::melt(distance_matrix2, na.rm = TRUE, value.name="Distance",varnames=c("Sample1","Sample2"), id.vars = "id")


  #label_color <- ifelse(meta_selected$Source[match(colnames(matrix_reduced),meta_selected$ID)] == "human", "blue", "black")
  #label_color <- ifelse(metadata[match(colnames(distance_matrix2),names(metadata))] == internal_name, "black", "blue")
  
  
  
  p1 <- ggplot(distance_matrix.m, aes(Sample1, Sample2)) +
    geom_tile(aes(fill = Distance), colour = "white", show.legend = F) + 
    geom_text(aes(label = round(Distance, 1)), size = param_size) +
    scale_fill_gradientn(colours = c("white", "yellow", "orange", "red"), values = scales::rescale(param_color_thr)) +
    theme(axis.text.x = element_text(angle = 90, vjust = 0.5, size = param_legendsize, hjust = 1, color = label_color),  axis.text.y = element_text(size = param_legendsize, color = label_color), axis.title.x = element_blank(),
          axis.title.y = element_blank()) + coord_fixed()
  
  
  return(p1)
  
}




# read in data -----------------------------------


distance_matrix <- distance_matrix_file %>% read_distance_matrix (sep = "", header = F)

samples_extern <- read.table(sample_list_file, sep ="\t", stringsAsFactors = F, header = T)$sample


# fallunterscheidung:
# only 1 external sample
# no hit below threshold

if(length(samples_extern) == 1){
  # select external in rows and internal in columns
  distance_matrix_selected <- distance_matrix[rownames(distance_matrix) %in% samples_extern,! rownames(distance_matrix) %in% samples_extern]
  
  
  # keep only internal samples below threshold
  distance_matrix_reduced <- t(as.matrix(distance_matrix_selected[distance_matrix_selected < threshold]))
  rownames(distance_matrix_reduced) <- samples_extern

  # action if no match
  if(ncol(distance_matrix_reduced) == 0) {
    warning("No matches between external and internal matches found below threshold")
    #distance_matrix_reduced <- NA
  }

} else { # more than 1 external sample
  
  # select external in rows and internal in columns
  distance_matrix_selected <- distance_matrix[rownames(distance_matrix) %in% samples_extern,! rownames(distance_matrix) %in% samples_extern] 
  
  # filter with threshold
  distance_matrix_reduced_list <- apply(distance_matrix_selected,2, function(col) {
    if(any(col < threshold)){
      return(col)
    }
  })
  
  
  # TODO check fix check if not a list but a matrix
  if(is.list(distance_matrix_reduced_list)){
    distance_matrix_reduced <- do.call(cbind,distance_matrix_reduced_list[sapply(distance_matrix_reduced_list,length) > 0])
    
  } else if (is.matrix(distance_matrix_reduced_list)){
    distance_matrix_reduced <- distance_matrix_reduced_list
  } else {
    warning("No matches between external and internal matches found below threshold")
    #distance_matrix_reduced <- distance_matrix[rownames(distance_matrix) %in% samples_extern, rownames(distance_matrix) %in% samples_extern]
    distance_matrix_reduced <- NA
  }
  
  
}

write.table(round(distance_matrix_reduced,0),file=distance_table_out,sep="\t", row.names = T, quote = F, col.names = NA)
write.table(t(round(distance_matrix_reduced,0)),file=distance_table_out2,sep="\t", row.names = T, quote = F, col.names = NA)

#p_distance_table <- plot_distance_matrix(distance_matrix_reduced,metadata = NA)
# ggsave(distance_table_pdf)



samples_close <- unique(c(colnames(distance_matrix_reduced),rownames(distance_matrix_reduced)))
write(samples_close, file = samples_related_file)

if (length(samples_close) > 1) { # if matrix as at least 2 elements

  distance_matrix_reduced2 <- distance_matrix[samples_close,samples_close]
  
  # cluster distance matrix
  distance_matrix_reduced2_clustered <- hclust(as.dist(distance_matrix_reduced2))
  distance_matrix_reduced2 <- distance_matrix_reduced2[distance_matrix_reduced2_clustered$order, distance_matrix_reduced2_clustered$order]
  
  distance_matrix_reduced2 <- round(distance_matrix_reduced2,0)
  
  
} else {
  print("no match")
  distance_matrix_reduced2 <- NA
}

# output ----------

#print(distance_matrix_reduced2)
write.table(distance_matrix_reduced2,file=distance_out,sep="\t", row.names = T, quote = F, col.names = NA)
samples_origin <- ifelse(samples_close %in% samples_extern,external_name,internal_name)
names(samples_origin) <- samples_close

write.table(cbind(id=samples_close,origin=samples_origin),file=sample_meta_file,sep="\t", row.names = F, quote = F, col.names = T)


# plot distance matrix ----------




# plot tree ----------


#round(distance_matrix_reduced2,0)

#distance_matrix_cluster <- round(distance_matrix[samples_cluster,samples_cluster],0)


if(is.matrix(distance_matrix_reduced2)){
  
  # export tsv
  distance_matrix_reduced2b <- reorder_cormat(distance_matrix_reduced2)

  # export pdf
  label_color <- ifelse(samples_origin[match(colnames(distance_matrix_reduced2),names(samples_origin))] == internal_name, "black", "blue")
  p_cluster <- plot_distance_matrix(distance_matrix_reduced2,label_color = label_color) #metadata = samples_origin, 
  ggsave(distance_matrix_pdf)

  # export as newick
  clust <- distance_matrix_reduced2 %>% as.dist %>% hclust(., method = clustering_method)
  ape::write.tree(as.phylo(clust), file = tree_nwk)
  
  # export tree as pdf and png
  
  # if(nrow(distance_matrix_reduced2) > 2){
  #   pdf(file = tree_pdf, width = 16  , height = 9 )
  #   #par(mar=c(2,2,4.1,6.1))
  #   plot(clust, main = paste("Single-linkage tree"), sub = "")
  #   dev.off()
  #} else {
  
    phylo <- as.phylo(clust)
    edge_color <- ifelse(phylo$edge.length<1,"white","red") 
    tip_color <- label_color #ifelse(samples_meta$origin[match(phylo$tip.label,samples_meta$id)]==internal_name,"black","blue")
    if (length(phylo$tip.label) > 50) {
      my_cex_tiplabel <- 0.5
      my_cex_edgelabel <- 0.5
    } else{
      my_cex_tiplabel <- 0.8
      my_cex_edgelabel <- 0.8
    }
    
    print("Export tree as pdf ...")
    pdf(file = tree_pdf, width = 16  , height = 9 )
    
    #plot(phylo,show.node.label = T, tip.color = 2)
    plot(phylo, main = paste0(clustering_method,"-linkage tree"), cex=my_cex_tiplabel, tip.color = tip_color)
    edgelabels(round(phylo$edge.length, 0), srt = 0, bg = "white", adj = c(0.5,1), col = edge_color, cex = my_cex_edgelabel, frame = "none")
    
    dev.off()
    #}

    # print("Export tree as png ...")
    # png(file = tree_png, width = 16  , height = 9 )
    # plot(phylo, main = paste0(clustering_method,"-linkage tree"), cex=my_cex_tiplabel, tip.color = tip_color)
    # edgelabels(round(phylo$edge.length, 0), srt = 0, bg = "white", adj = c(0.5,1), col = edge_color, cex = my_cex_edgelabel, frame = "none")
    # dev.off()
    
    
    
    
}

# ----------

print("Output files")
print(paste("distance_out:",distance_out))
print(paste("distance_table_out:",distance_table_out))
print(paste("samples_related_file:",samples_related_file))
print(paste("sample_meta_file:",sample_meta_file))



# # top 10 hits
# apply(distance_matrix_selected,1,function(row){
#   head(sort(row),10)
# })
# 
# # best hit


write(paste("Summary of matches\n"), file = match_overview_file, append = F)


if (length(samples_extern) > 1){
  


  for (row in 1:nrow(distance_matrix_selected)){
    mymin <- min(distance_matrix_selected[row,])
    print(paste("query Sample", rownames(distance_matrix_selected)[row],":"))
    write(paste("query Sample", rownames(distance_matrix_selected)[row],":"), file = match_overview_file, append = T)
    matches <- which(distance_matrix_selected[row,] < 30)
    distances <- distance_matrix_selected[row,matches]
    subjects <- colnames(distance_matrix_selected)[matches]
    if(length(distances) > 0){
      print(paste("match:",subjects,distances))
      write(paste("match:",subjects,distances), file = match_overview_file, append = T)
    }
  }
} else {

  mymin <- min(distance_matrix_selected)
  print(paste("query Sample", rownames(distance_matrix_selected),":"))
  write(paste("query Sample", rownames(distance_matrix_selected),":"), file = match_overview_file, append = T)
  matches <- which(distance_matrix_selected < 30)
  distances <- distance_matrix_selected[matches]
  subjects <- names(distances) #names(distance_matrix_selected)[matches]
  if(length(distances) > 0){
    print(paste("match:",subjects,distances))
    write(paste("match:",subjects,distances), file = match_overview_file, append = T)
  }
  
  
  
    
}  
