#!/bin/bash

# NOTE: This script requires a test result directory name prefix as an argument
# e.g. bash chewieSnake_test.sh "test_2020-12-24"

#region ########## Setup

CORES=7        # Workstations: 7 ; Cluster: 40
CORES_SAMPLE=3 # Workstations: 3 ; Cluster: 10

function define_paths() {
  dir_chewieSnake=$(dirname $(dirname $(readlink -f $0))) # /home/carlus/BfR/Snakemake/chewieSnake
  dir_data="${dir_chewieSnake}/testdata/$1" # for chewieSnake
  dir_schema=${dir_chewieSnake}/schema/enterobase_senterica_cgmlst
  dir_prodigal_training=${dir_chewieSnake}/chewBBACA/CHEWBBACA/prodigal_training_files/Salmonella_enterica.trn
  dir_analysis_docker=/chewieSnake/analysis
  dir_schema_docker=/chewieSnake/schema/enterobase_senterica_cgmlst
  dir_prodigal_training_docker=/chewieSnake/chewBBACA/CHEWBBACA/prodigal_training_files/Salmonella_enterica.trn
}

function init_conda() {
  CONDA_BASE=$(conda info --base)
  CONDA_ENV=$(head -1 ${dir_chewieSnake}/envs/chewiesnake.yaml | cut -d' ' -f2) # TODO: eventually replace with chewiesnake_fromconda.yaml

  if [[ $CONDA_BASE != */envs/${CONDA_ENV}* ]]; then
    echo "Conda environment $CONDA_ENV has not been activated yet. Activating Conda ..."
    source $CONDA_BASE/etc/profile.d/conda.sh
    conda activate $CONDA_ENV
  else
    CONDA_BASE=$(dirname $(dirname $CONDA_BASE))
    echo "Conda environment $CONDA_ENV was already activated yet. Setting CONDA_BASE to $CONDA_BASE ..."
  fi
}

function create_test_dir() {
  if [ -z "$1" ]; then
    dir_test="test_$(date +"%Y%m%d%H%M")"
  else
    dir_test=$1
  fi

  if [ ! -d "${dir_data}/${dir_test}" ]; then
    mkdir -p ${dir_data}/${dir_test}
    echo "Creating Test Directory: ${dir_data}/${dir_test} ..."
  else
    echo "Using Test Directory: ${dir_data}/${dir_test} ..."
  fi
  echo "Directory changed to ${dir_data}/${dir_test} ..."
  cd ${dir_data}/${dir_test}
}

#endregion
#region ########## Test Section: ChewieSnake Docker

chewieMode="chewiesnake"
define_paths "$chewieMode"

docker pull bfrbioinformatics/chewiesnake:latest || true
docker_image_id=$(docker image list | grep "chewiesnake" | grep "latest" | awk '{ print $3 }')
create_test_dir "$1_docker" # get dir name from script argument (PyCharm)
cp -f -r ${dir_data}/data ${dir_data}/${dir_test}

docker run --rm \
  -v ${dir_data}/${dir_test}:${dir_analysis_docker} \
  --entrypoint bash \
  $docker_image_id \
  /chewieSnake/scripts/create_sampleSheet.sh \
  --force \
  --mode assembly \
  --fastxDir ${dir_analysis_docker}/data \
  --outDir ${dir_analysis_docker}/data

docker run --rm \
  -v ${dir_data}/${dir_test}:${dir_analysis_docker} \
  -e LOCAL_USER_ID=$(id -u $USER) \
  $docker_image_id \
  --condaprefix /opt/conda/envs \
  --sample_list ${dir_analysis_docker}/data/samples.tsv \
  --working_directory ${dir_analysis_docker} \
  --scheme ${dir_schema_docker} \
  --prodigal ${dir_prodigal_training_docker} \
  --threads $CORES
#  --dryrun \

#endregion
#region ########## Test Section: ChewieSnake

chewieMode="chewiesnake"
define_paths "$chewieMode"
init_conda

# full conda test via wrapper: 7 min 30 sec on 7 threads

create_test_dir "$1_wrapper" # get dir name from script argument (PyCharm)
ln --force --symbolic ${dir_data}/data ${dir_data}/${dir_test}/data
${dir_chewieSnake}/chewieSnake.py \
  --sample_list ${dir_data}/${dir_test}/data/samples.tsv \
  --working_directory ${dir_data}/${dir_test} \
  --scheme ${dir_schema} \
  --prodigal ${dir_prodigal_training} \
  --threads $CORES
#  --dryrun \

# export dag of "$1_wrapper"

snakemake \
  --snakefile ${dir_chewieSnake}/chewieSnake_fromcontigs.smk \
  --configfile ${dir_data}/${dir_test}/config.yaml \
  --forceall \
  --cores $CORES \
  --dag | dot -Tpdf > ${dir_data}/${dir_test}/dag.pdf

# full conda test via config.yaml

create_test_dir "$1_config" # get dir name from script argument (PyCharm)
ln --force --symbolic ${dir_data}/data ${dir_data}/${dir_test}/data
sed -E 's@_wrapper@_config@g' ${dir_data}/${dir_test/"_config"/"_wrapper"}/config.yaml > ${dir_data}/${dir_test}/config.yaml
snakemake \
  --keep-going \
  --configfile ${dir_data}/${dir_test}/config.yaml \
  --snakefile ${dir_chewieSnake}/chewieSnake_fromcontigs.smk \
  --cores $CORES
#  --dryrun \

#endregion
#region ########## Test Section: Allele Hashing

# Single Allele Hashing for result set "$1_config"

${dir_chewieSnake}/scripts/alleleprofile_hasher.py \
  --profile ${dir_data}/${dir_test}/cgmlst/GMI15-002-DNA/allele_profiles.fsremoved.tsv \
  --database ${dir_schema} \
  --outfile ${dir_data}/${dir_test}/allele_hashing/GMI15-002-DNA_allele_profiles.tsv \
  --force

# Multi Allele Hashing for result set "$1_config"

${dir_chewieSnake}/scripts/alleleprofile_hasher.py \
  --profile ${dir_data}/${dir_test}/cgmlst/allele_profiles_unhashed.tsv \
  --database ${dir_schema} \
  --outfile ${dir_data}/${dir_test}/allele_hashing/multi_allele_profiles.tsv \
  --force

#endregion
#region ########## Test Section: Frameshift

create_test_dir "$1_frameshifts" # get dir name from script argument (PyCharm)
ln --force --symbolic ${dir_data}/data ${dir_data}/${dir_test}/data
${dir_chewieSnake}/chewieSnake.py \
  --sample_list ${dir_data}/${dir_test}/data/samples.tsv \
  --working_directory ${dir_data}/${dir_test} \
  --scheme ${dir_schema} \
  --prodigal ${dir_prodigal_training} \
  --remove_frameshifts \
  --threads $CORES
#  --dryrun \

#endregion
#region ########## Test Section: Comparison

create_test_dir "$1_comparison" # get dir name from script argument (PyCharm)
ln --force --symbolic ${dir_data}/data ${dir_data}/${dir_test}/data
${dir_chewieSnake}/chewieSnake.py \
  --sample_list ${dir_data}/${dir_test}/data/samples.tsv \
  --working_directory ${dir_data}/${dir_test} \
  --scheme ${dir_schema} \
  --prodigal ${dir_prodigal_training} \
  --comparison \
  --comparison_db ${dir_data}/${dir_test}/data/comparison_allele_profiles.tsv \
  --threads $CORES
#  --dryrun \

#endregion
#region ########## Test Section: ChewieSnake_Join

chewieMode="chewiesnake_join"
define_paths "$chewieMode"
init_conda

# Prerequisite: prepare dataset
# * need allele profiles, allele stats and timestamps from at least 2 partners
# * need to rename samples
# * works with empty allele_profiles?

create_test_dir "$1_joinedDB"
ln --force --symbolic ${dir_data}/data ${dir_data}/${dir_test}/data
${dir_chewieSnake}/chewieSnake_join.py \
  --sample_list ${dir_data}/${dir_test}/data/samples.tsv \
  --working_directory ${dir_data}/${dir_test} \
  --serovar_info ${dir_data}/sample_serovars.tsv \
  --external_cluster_names ${dir_data}/epiclusters.tsv \
  --threads $CORES
#  --dryrun \

# export dag of "$1_joinedDB"

snakemake \
  --snakefile ${dir_chewieSnake}/chewieSnake_join.smk \
  --configfile ${dir_data}/${dir_test}/config.yaml \
  --forcerun \
  --cores $CORES \
  --dag | dot -Tpdf >${dir_data}/${dir_test}/dag.pdf

# Notes from @carlus :
# -e EXTERNAL_CLUSTER_NAMES OR --external_cluster_names EXTERNAL_CLUSTER_NAMES
# HELP: A tab seperated file with external cluster names, e.g. Epi clusers. Must contain one representative sample per cluster: sample <tab >cluster_name
#
# --cluster_names
# HELP: A tab seperated file with serovar or clade names: sample <tab >cluster_name
#
# --cluster
# HELP:

#endregion
