#!/usr/bin/env Rscript

# GOAL: read in clusters and provide summary

# define ----

#debug <- T
debug <- F

if (debug){

  sender_info_file  <- file.path(workdir, "merged_db/sender_sample_info.tsv")
  sample_timestamps_file <- file.path(workdir, "merged_db/timestamps.tsv")

  mapping_sample2clustername_file <- file.path(workdir, "merged_db","mapping_sample2clustername.tsv")
  mapping_sample2externalclustername_file = file.path(workdir, "merged_db/mapping_sample2externalclustername.tsv")
  
  # not needed:
  orphansamples_file = file.path(workdir, "merged_db/orphan_samples.tsv")

  
  samples_in_clusters_file <- file.path(workdir, "merged_db", "sample_clustering.tsv")
} else { 
# snakemake input ----------------------------------------

  workdir <- snakemake@params[["workdir"]]    
  
# input
sender_info_file  <- file.path(workdir, snakemake@input[["sender_info"]])
sample_timestamps_file <- file.path(workdir, snakemake@input[["timestamps"]])

mapping_sample2clustername_file <- file.path(workdir, snakemake@input[["mapping_sample2clustername_file"]])
mapping_sample2externalclustername_file <- file.path(workdir, snakemake@input[["mapping_sample2externalclustername_file"]])

mapping_sample2serovar.file <- snakemake@input[["mapping_sample2serovar"]]

orphansamples_file <- file.path(workdir, snakemake@input[["orphansamples_file"]])

# subcluster_files
subcluster_files <- file.path(workdir, snakemake@input[["subclusters"]])

# output

samples_in_clusters_file <- file.path(workdir, snakemake@output[["clustering_info"]])
samples_orphans_file <- file.path(workdir, snakemake@output[["orphan_info"]])



}

# load data -----------------------------------------------------

# load timing data
sample_timestamps <- read.delim(sample_timestamps_file, stringsAsFactors = F, header = F, sep = "\t", col.names = c("Sample","Timestamp"))

# load sender info
sender_info <- read.delim(sender_info_file, stringsAsFactors = F, header = T, sep = "\t", col.names = c("Sample","Sender"))


samples_in_clusters <- read.delim(file = mapping_sample2clustername_file, sep="\t", stringsAsFactors = F, header = T)

print("subcluster_files:")
print(subcluster_files)

subclusters <- do.call(rbind,lapply(subcluster_files, function(x) read.delim(x, stringsAsFactors = F) ) )

# example
#subcluster_files <- "/home/carlus/BfR/Projects/salmonella-alleledb/merged_db/clustering/Archimelus/mapping_sample2subclustername.tsv"
#subclusters <- read.delim(subcluster_file, stringsAsFactors = F)

# epi clusters
mapping_sample2externalclustername <- read.delim(file = mapping_sample2externalclustername_file, sep="\t", stringsAsFactors = F, header = T)
colnames(mapping_sample2externalclustername) <- sub("representative_sample","representative_sample_externalcluster",colnames(mapping_sample2externalclustername))
mapping_sample2externalclustername <- mapping_sample2externalclustername[,colnames(mapping_sample2externalclustername) != "cluster_name"]

# orphan samples
orphansamples <- read.delim(orphansamples_file, stringsAsFactors = F)

# sample names of all samples in data set
samples <- sender_info$Sample

# execution -------------------------------------------------------------

# # initialize samples in cluster data frame
#samples_in_clusters <- clusters_membership

samples_in_clusters <- merge(samples_in_clusters,sender_info, by.x = "sample", by.y = "Sample")


# merge with date
samples_in_clusters <- merge(x=samples_in_clusters,y=sample_timestamps, by.x = "sample", by.y = "Sample", all.x = T, all.y = F)
samples_in_clusters$Timestamp <- as.Date(samples_in_clusters$Timestamp)

# define representative (=oldest) sample per cluster
cluster_names_uniq <- unique(samples_in_clusters$cluster_name)

Clusters_representativesample <- sapply(cluster_names_uniq, function(cluster_name){
  temp <- samples_in_clusters[samples_in_clusters$cluster_name == cluster_name,]
  # find oldest
  oldest_date <- temp$Timestamp[order(temp$Timestamp)][1]
  oldest_samples <- which(temp$Timestamp == oldest_date)
  if(length(oldest_samples) == 1){
    return(temp$sample[oldest_samples])
  } else {
    # return first in alphabetical order
    return(sort(temp$sample[oldest_samples])[1])
  }
})

# convert to dataframe
Clusters_representativesample_df <- data.frame(representative_sample=Clusters_representativesample, cluster_name = names(Clusters_representativesample))

# merge representative name with samples in clusters
samples_in_clusters <- merge(x = samples_in_clusters, y = Clusters_representativesample_df, by.x = "cluster_name", by.y = "cluster_name", all.x = T)

# add epi clusters
samples_in_clusters <- merge(x = samples_in_clusters, y = mapping_sample2externalclustername, by.x = "sample", by.y = "sample", all.x = T, all.y = F)

# add serovar if available
if(!is.null(mapping_sample2serovar.file)){ # if option is not set value is NULL
  if (file.info(mapping_sample2serovar.file)$size != 0) {
    print("Match serovar info to cluster")
    mapping_sample2serovar <- read.delim(mapping_sample2serovar.file, stringsAsFactors = F)
    # rename column
    colnames(mapping_sample2serovar)[colnames(mapping_sample2serovar) == "external_cluster_name"] <- "serovar"
    samples_in_clusters <- merge(x = samples_in_clusters, y = mapping_sample2serovar[,c("sample","serovar")], by.x = "sample", by.y = "sample", all.x = T, all.y = F)
  }
}

# add subclusters ----------------
samples_in_clusters <- merge(x = samples_in_clusters, y = subclusters, by.x = "sample", by.y = "sample", all.x = T, all.y = T)
samples_in_clusters$cluster_code <- apply(samples_in_clusters[,grep("^cluster_name",colnames(samples_in_clusters), value = F)], 1, function(x) gsub(" ","",paste0(x,collapse = ".")))


# reorder columns, Sample is first column
samples_in_clusters <- samples_in_clusters[,c(which(colnames(samples_in_clusters) == "sample"),which(colnames(samples_in_clusters) != "sample"))]

# orphan analysis -----------------
#samples_orphans <- data.frame(sample = samples[which(! samples %in% samples_in_clusters$sample)])
#samples_orphans$cluster_name <- paste("Orphan",1:nrow(samples_orphans), sep = "_")

samples_orphans <- orphansamples
# sender info
samples_orphans <- merge(samples_orphans,sender_info, by.x = "sample", by.y = "Sample")
# timestamp info
samples_orphans <- merge(x=samples_orphans,y=sample_timestamps, by.x = "sample", by.y = "Sample", all.x = T, all.y = F)
samples_orphans$Timestamp <- as.Date(samples_orphans$Timestamp)

# export samples_in_clusters ----------------

write.table(samples_in_clusters, file = samples_in_clusters_file, sep="\t", quote = F, row.names = F)

# export orphans?

write.table(samples_orphans, file = samples_orphans_file, sep="\t", quote = F, row.names = F)


