# clustering and cluster naming

# libraries -----------------------------------

suppressPackageStartupMessages(library(magrittr))
suppressPackageStartupMessages(library(dendextend))
suppressPackageStartupMessages(library(reshape2)) # for converting matrix
suppressPackageStartupMessages(library(dplyr))

# functions -----------------------------------
#source("scripts/functions.R") # see below

# define data -----------------------------------
#debug <- T # load parameters manually
debug <- F

if(debug){
  source("parameters_clustering.R")  
} else {
  #  wd <- snakemake@params[["workdir"]]  
  
  # input
  
  # allele distances
  distance_matrix_file <- snakemake@input[["distance_matrix_tsv"]]

  # clustername_sample_association
  mapping_sample2clustername_file <- snakemake@input[["mapping_sample2clustername_file"]]
  
  # output
  orphansamples_file <- snakemake@output[["orphansamples_file"]]
  
  #mapping_sample2clustername_fileout <- snakemake@output[["mapping_sample2clustername_file"]]
  clustering_dir <- snakemake@output[["clustering_dir"]]

  mapping_sample2clustername_fileout <- snakemake@params[["mapping_sample2clustername_file"]]
  #clustering_dir <- snakemake@params[["clustering_dir"]]
  
  
  # parameters
  clustering_method <- snakemake@params[["clustering_method"]]  
  cut_threshold <- snakemake@params[["distance_threshold"]]  
  # cluster name list
  potential_names_file <- snakemake@params[["potential_names_file"]]  

  # functions file
  #functions_file <- snakemake@params[["functions_file"]]
  functions_file <- file.path(snakemake@params[["functions_file"]],"scripts/functions.R")
  
  
  log <- file(snakemake@log[[1]], open="wt")
  
  # redirect output to log file
  sink(log)
  sink(log, type="message")

  

  # check file existence
  if(! file.exists(distance_matrix_file)) stop("File",distance_matrix_file,"does NOT exist")
  
  if(! file.exists(mapping_sample2clustername_file)) stop("File",mapping_sample2clustername_file,"does NOT exist")
  
  if(! file.exists(potential_names_file)) stop("File",potential_names_file,"does NOT exist")
  
  print("All files exist")
    
}

dir.create(clustering_dir,showWarnings = F)

# load functions ----
print(paste("File to functions.R",functions_file))

if(! file.exists(functions_file)) stop("Function file",functions_file,"does NOT exist")

source(functions_file)

# load data -----------------------------------
distance_matrix <- read_distance_matrix(distance_matrix_file)

mapping_sample2clustername_loaded <- read.delim(file = mapping_sample2clustername_file, sep="\t", stringsAsFactors = F, header = T, colClasses = "character")
mapping_sample2clustername <- mapping_sample2clustername_loaded

if(!file.exists(potential_names_file)) stop("File with potential cluster names",potential_names_file,"does NOT exist\nExpected a text file with one name per row")
potential_clusternames <- readLines(potential_names_file)

print("All files loaded")

# execution -----------------------------------
# Clustering
clustering <- distance_matrix %>% as.dist %>% hclust(., method = clustering_method) %>% as.dendrogram

# cluster numbers
samples2cluster <- Get_samples2cluster(dendro = clustering, cut_threshold = cut_threshold, removeOrphans = T)

#list of orphans
samples2cluster_orphans <- Get_samples2cluster(dendro = clustering, cut_threshold = cut_threshold, removeOrphans = F) %>% filter(grepl("Orphan",cluster_number))


#TEST simulation of new entries
#samples2cluster_new <- data.frame(sample=c("new2_38_1","new2_38_2","new2_38_3","new2_88_1","new2_88_2","new2_100_1","new2_100_2"), cluster_number = c("38","38","38","88","88","100","100"))
#samples2cluster <- rbind(samples2cluster,samples2cluster_new)

# match samples to existing cluster names
samples2clustername <- merge(x=samples2cluster,y=mapping_sample2clustername, by.x="sample", by.y = "sample", all.x = T, all.y = F)

# find all unique cluster_number to cluster_name pairs
clusternumber2name <- samples2clustername[,c("cluster_number","cluster_name")] %>% filter(!is.na(cluster_name)) %>% unique

# check cluster merge and split conflicts
check_clusternaming_conflicts(clusternumber2name)

# Fix potential merge conflicts
after_mergeconflict_fixing <- Fix_merge_conflicts (clusternumber2name,samples2clustername)
clusternumber2name <- after_mergeconflict_fixing$clusternumber2name
samples2clustername <- after_mergeconflict_fixing$samples2clustername

# add new samples with same cluster number
update_results <- update_samplenames (samples2clustername, clusternumber2name, verbose=T, potential_clusternames = potential_clusternames)
samples2clustername <- update_results$samples2clustername
clusternumber2name <- update_results$clusternumber2name

mapping_sample2clustername_updated <- samples2clustername[,c("sample","cluster_name")]

# save results -----------------

print("Saving results")


print("write updated mapping_sample2clustername_fileout")
# write updated mapping_sample2clustername_fileout
write.table(mapping_sample2clustername_updated, file = mapping_sample2clustername_fileout, sep="\t", quote = F, row.names = F)

print("write orphan list")
# write orphan list
write.table(samples2cluster_orphans, file = orphansamples_file, sep="\t", quote = F, row.names = F)

# samples2clustername
# 1 file per cluster ?
print("Writing per cluster files")
cluster_names_uniq <- unique(clusternumber2name$cluster_name)

for(cluster in cluster_names_uniq){
  mycluster_dir <- file.path(clustering_dir, cluster) 
  print(paste("Writing",mycluster_dir))
  #dir.create(mycluster_dir, showWarnings = F)
  # subset file
  #mycluster_file <- file.path(mycluster_dir,"samples2clustername.tsv")
  mycluster_file <- file.path(clustering_dir,paste0("samples_",cluster,".tsv"))
  print(mycluster_file)
  samples2clustername_cluster <- samples2clustername %>% filter(cluster_name == cluster)
  write.table(samples2clustername_cluster, file = mycluster_file, sep="\t", quote = F, row.names = F)
}
#


print("Done")

#TODO ---------------------
# * add cluster creation date
# * create cluster hash from creation date
# Need to load

# as.numeric(format(Sys.time(), "%OS3")) * 1000
# 
# my_time <- Sys.time()
# 
# format(my_time, "%Y-%m-%d")
# format(my_time, "%H:%M:%OS3")
# 
# for(i in 1:10){
#   Sys.sleep(0.01)
#   print(format(Sys.time(), "%H:%M:%OS3"))
# }
# 
# for(i in 1:10){
#   Sys.sleep(0.01)
#   print(as.numeric(format(Sys.time(), "%OS3")) * 1000)
# }
